import sys

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from cleo.exceptions import LogicException

from .argument import Argument
from .option import Option


class Definition:
    """
    A Definition represents a set of command line arguments and options.
    """

    def __init__(
        self, definition: Optional[List[Union[Argument, Option]]] = None
    ) -> None:
        self._arguments: Dict[str, Argument] = {}
        self._required_count = 0
        self._has_a_list_argument = False
        self._has_optional = False
        self._options: Dict[str, Option] = {}
        self._shortcuts: Dict[str, str] = {}

        if definition is None:
            definition = []

        self.set_definition(definition)

    @property
    def arguments(self) -> List[Argument]:
        return list(self._arguments.values())

    @property
    def argument_count(self) -> int:
        if self._has_a_list_argument:
            return sys.maxsize

        return len(self._arguments)

    @property
    def required_argument_count(self) -> int:
        return self._required_count

    @property
    def argument_defaults(self) -> Dict[str, Any]:
        values = {}

        for argument in self._arguments.values():
            values[argument.name] = argument.default

        return values

    @property
    def options(self) -> List[Option]:
        return list(self._options.values())

    @property
    def option_defaults(self) -> Dict[str, Any]:
        values = {}
        for option in self._options.values():
            values[option.name] = option.default

        return values

    def set_definition(self, definition: List[Union[Argument, Option]]) -> None:
        arguments = []
        options = []

        for item in definition:
            if isinstance(item, Option):
                options.append(item)
            else:
                arguments.append(item)

        self.set_arguments(arguments)
        self.set_options(options)

    def set_arguments(self, arguments: List[Argument]) -> None:
        self._arguments = {}
        self._required_count = 0
        self._has_a_list_argument = False
        self._has_optional = False
        self.add_arguments(arguments)

    def add_arguments(self, arguments: List[Argument]) -> None:
        for argument in arguments:
            self.add_argument(argument)

    def add_argument(self, argument: Argument) -> None:
        if argument.name in self._arguments:
            raise LogicException(
                f'An argument with name "{argument.name}" already exists'
            )

        if self._has_a_list_argument:
            raise LogicException("Cannot add an argument after a list argument")

        if argument.is_required() and self._has_optional:
            raise LogicException("Cannot add a required argument after an optional one")

        if argument.is_list():
            self._has_a_list_argument = True

        if argument.is_required():
            self._required_count += 1
        else:
            self._has_optional = True

        self._arguments[argument.name] = argument

    def argument(self, name: Union[str, int]) -> Argument:
        if not self.has_argument(name):
            raise ValueError(f'The "{name}" argument does not exist')

        if isinstance(name, int):
            arguments = list(self._arguments.values())
        else:
            arguments = self._arguments

        return arguments[name]

    def has_argument(self, name: Union[str, int]) -> bool:
        if isinstance(name, int):
            arguments = list(self._arguments.values())
        else:
            arguments = self._arguments

        try:
            arguments[name]
        except (KeyError, IndexError):
            return False

        return True

    def set_options(self, options: List[Option]) -> None:
        self._options = {}
        self._shortcuts = {}
        self.add_options(options)

    def add_options(self, options: List[Option]) -> None:
        for option in options:
            self.add_option(option)

    def add_option(self, option: Option) -> None:
        if option.name in self._options and option != self._options[option.name]:
            raise LogicException(f'An option named "{option.name}" already exists')

        if option.shortcut:
            for shortcut in option.shortcut.split("|"):
                if shortcut in self._shortcuts and option != self._shortcuts[shortcut]:
                    raise LogicException(
                        f'An option with shortcut "{shortcut}" already exists'
                    )

        self._options[option.name] = option

        if option.shortcut:
            for shortcut in option.shortcut.split("|"):
                self._shortcuts[shortcut] = option.name

    def option(self, name: str) -> Option:
        if not self.has_option(name):
            raise ValueError(f'The option "--{name}" option does not exist')

        return self._options[name]

    def has_option(self, name: str) -> bool:
        return name in self._options

    def has_shortcut(self, shortcut: str) -> bool:
        return shortcut in self._shortcuts

    def option_for_shortcut(self, shortcut: str) -> Option:
        return self._options[self.shortcut_to_name(shortcut)]

    def shortcut_to_name(self, shortcut: str) -> str:
        if shortcut not in self._shortcuts:
            raise ValueError(f'The "-{shortcut}" option does not exist')

        return self._shortcuts[shortcut]

    def synopsis(self, short: bool = False) -> str:
        elements = []

        if short and self._options:
            elements.append("[options]")
        elif not short:
            for option in self._options.values():
                value = ""
                if option.accepts_value():
                    value = " {}{}{}".format(
                        "[" if not option.requires_value() else "",
                        option.name.upper(),
                        "]" if not option.requires_value() else "",
                    )

                shortcut = ""
                if option.shortcut:
                    shortcut = "-{}|".format(option.shortcut)

                elements.append("[{}--{}{}]".format(shortcut, option.name, value))

        if elements and self._arguments:
            elements.append("[--]")

        tail = ""
        for argument in self._arguments.values():
            element = f"<{argument.name}>"
            if argument.is_list():
                element += "..."

            if not argument.is_required():
                element = "[" + element
                tail += "]"

            elements.append(element)

        return " ".join(elements) + tail



from typing import Optional
from typing import cast

from .inputs.input import Input
from .io import IO
from .outputs.buffered_output import BufferedOutput


class BufferedIO(IO):
    def __init__(
        self,
        input: Optional[Input] = None,
        decorated: bool = False,
        supports_utf8: bool = True,
    ) -> None:
        super(BufferedIO, self).__init__(
            input,
            BufferedOutput(decorated=decorated, supports_utf8=supports_utf8),
            BufferedOutput(decorated=decorated, supports_utf8=supports_utf8),
        )

        self._output = cast(BufferedOutput, self._output)
        self._error_output = cast(BufferedOutput, self._error_output)

    def fetch_output(self) -> str:
        return self._output.fetch()

    def fetch_error(self) -> str:
        return self._error_output.fetch()

    def clear(self) -> None:
        self._output.clear()
        self._error_output.clear()

    def clear_output(self) -> None:
        self._output.clear()

    def clear_error(self) -> None:
        self._error_output.clear()

    def supports_utf8(self) -> bool:
        return self._output.supports_utf8()

    def clear_user_input(self) -> None:
        self._input.stream.truncate(0)
        self._input.stream.seek(0)

    def set_user_input(self, user_input: str) -> None:
        self.clear_user_input()

        self._input.stream.write(user_input)
        self._input.stream.seek(0)



import inspect
import re

from typing import TYPE_CHECKING
from typing import Any
from typing import List
from typing import Optional
from typing import Union

from cleo import exceptions
from cleo.cursor import Cursor
from cleo.io.inputs.string_input import StringInput
from cleo.io.io import IO
from cleo.io.null_io import NullIO
from cleo.io.outputs.output import Verbosity
from cleo.parser import Parser

from .base_command import BaseCommand


if TYPE_CHECKING:
    from cleo.ui.progress_bar import ProgressBar
    from cleo.ui.progress_indicator import ProgressIndicator
    from cleo.ui.question import Question


class Command(BaseCommand):

    arguments = []
    options = []

    aliases = []

    usages = []

    commands = []

    def __init__(self):
        self._io: Optional[IO] = None
        super(Command, self).__init__()

    @property
    def io(self):  # type: () -> IO
        return self._io

    def configure(self) -> None:
        if not self.name:
            doc = self.__doc__

            if not doc:
                for base in inspect.getmro(self.__class__):
                    if base.__doc__ is not None:
                        doc = base.__doc__
                        break

            if doc:
                self._parse_doc(doc)

        for argument in self.arguments:
            self._definition.add_argument(argument)

        for option in self.options:
            self._definition.add_option(option)

    def _parse_doc(self, doc):
        doc = doc.strip().split("\n", 1)
        if len(doc) > 1:
            self.description = doc[0].strip()
            signature = re.sub(r"\s{2,}", " ", doc[1].strip())
            definition = Parser.parse(signature)
            self.name = definition["name"]

            for argument in definition["arguments"]:
                self._definition.add_argument(argument)

            for option in definition["options"]:
                self._definition.add_option(option)
        else:
            self.description = doc[0].strip()

    def execute(self, io: IO) -> int:
        self._io = io

        try:
            return self.handle()
        except KeyboardInterrupt:
            return 1

    def handle(self) -> int:
        """
        Executes the command.
        """
        raise NotImplementedError()

    def call(self, name: str, args: Optional[str] = None) -> int:
        """
        Call another command.
        """
        if args is None:
            args = ""

        input = StringInput(args)
        command = self.application.get(name)

        return self.application._run_command(command, self._io.with_input(input))

    def call_silent(self, name, args=None):  # type: (str, Optional[str]) -> int
        """
        Call another command silently.
        """
        if args is None:
            args = ""

        args = StringInput(args)
        command = self.application.get(name)

        return self.application._run_command(command, NullIO(input))

    def argument(self, name: str):
        """
        Get the value of a command argument.
        """
        return self._io.input.argument(name)

    def option(self, name: str):
        """
        Get the value of a command option.
        """
        return self._io.input.option(name)

    def confirm(
        self, question: str, default: bool = False, true_answer_regex: str = "(?i)^y"
    ) -> bool:
        """
        Confirm a question with the user.
        """
        from cleo.ui.confirmation_question import ConfirmationQuestion

        question = ConfirmationQuestion(
            question, default=default, true_answer_regex=true_answer_regex
        )

        return question.ask(self._io)

    def ask(
        self, question: Union[str, "Question"], default: Optional[Any] = None
    ) -> Any:
        """
        Prompt the user for input.
        """
        from cleo.ui.question import Question

        if not isinstance(question, Question):
            question = Question(question, default=default)

        return question.ask(self._io)

    def secret(
        self, question: Union[str, "Question"], default: Optional[Any] = None
    ) -> Any:
        """
        Prompt the user for input but hide the answer from the console.
        """
        from cleo.ui.question import Question

        if not isinstance(question, Question):
            question = Question(question)

        question.hide()

        return question.ask(self._io)

    def choice(
        self,
        question: str,
        choices: List[str],
        default: Optional[Any] = None,
        attempts: Optional[int] = None,
        multiple: bool = False,
    ) -> Any:
        """
        Give the user a single choice from an list of answers.
        """
        from cleo.ui.choice_question import ChoiceQuestion

        question = ChoiceQuestion(question, choices, default)

        question.set_max_attempts(attempts)
        question.set_multi_select(multiple)

        return question.ask(self._io)

    def create_question(self, question, type=None, **kwargs):
        """
        Returns a Question of specified type.
        """
        from cleo.ui.choice_question import ChoiceQuestion
        from cleo.ui.confirmation_question import ConfirmationQuestion
        from cleo.ui.question import Question

        if not type:
            return Question(question, **kwargs)

        if type == "choice":
            return ChoiceQuestion(question, **kwargs)

        if type == "confirmation":
            return ConfirmationQuestion(question, **kwargs)

    def table(self, header=None, rows=None, style=None):
        """
        Return a Table instance.
        """
        from cleo.ui.table import Table

        table = Table(self._io, style=style)

        if header:
            table.set_headers([header])

        if rows:
            table.set_rows(rows)

        return table

    def render_table(self, headers, rows, style=None):
        """
        Format input to textual table.
        """
        table = self.table(headers, rows, style)

        table.render()

    def write(self, text, style=None):
        """
        Writes a string without a new line.
        Useful if you want to use overwrite().
        """
        if style:
            styled = "<%s>%s</>" % (style, text)
        else:
            styled = text

        self._io.write(styled)

    def line(
        self,
        text: str,
        style: Optional[str] = None,
        verbosity: Verbosity = Verbosity.NORMAL,
    ):
        """
        Write a string as information output.
        """
        if style:
            styled = "<{}>{}</>".format(style, text)
        else:
            styled = text

        self._io.write_line(styled, verbosity=verbosity)

    def line_error(
        self,
        text: str,
        style: Optional[str] = None,
        verbosity: Verbosity = Verbosity.NORMAL,
    ):
        """
        Write a string as information output to stderr.
        """
        if style:
            styled = "<%s>%s</>" % (style, text)
        else:
            styled = text

        self._io.write_error_line(styled, verbosity)

    def info(self, text):
        """
        Write a string as information output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "info")

    def comment(self, text):
        """
        Write a string as comment output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "comment")

    def question(self, text):
        """
        Write a string as question output.

        :param text: The line to write
        :type text: str
        """
        self.line(text, "question")

    def progress_bar(self, max: int = 0) -> "ProgressBar":
        """
        Creates a new progress bar
        """
        from cleo.ui.progress_bar import ProgressBar

        return ProgressBar(self._io, max=max)

    def progress_indicator(
        self,
        fmt: Optional[str] = None,
        interval: int = 100,
        values: Optional[List[str]] = None,
    ) -> "ProgressIndicator":
        """
        Creates a new progress indicator.
        """
        from cleo.ui.progress_indicator import ProgressIndicator

        return ProgressIndicator(self.io, fmt, interval, values)

    def spin(self, start_message, end_message, fmt=None, interval=100, values=None):
        """
        Automatically spin a progress indicator.
        """
        spinner = ProgressIndicator(self.io, fmt, interval, values)

        return spinner.auto(start_message, end_message)

    def add_style(self, name, fg=None, bg=None, options=None):
        """
        Adds a new style
        """
        style = Style(name)
        if fg is not None:
            style.fg(fg)

        if bg is not None:
            style.bg(bg)

        if options is not None:
            if "bold" in options:
                style.bold()

            if "underline" in options:
                style.underlined()

        self._io.output.formatter.add_style(style)
        self._io.error_output.formatter.add_style(style)

    def overwrite(self, text):
        """
        Overwrites the current line.

        It will not add a new line so use line('')
        if necessary.
        """
        self._io.overwrite(text)



from typing import Optional


class TableCellStyle:
    def __init__(
        self, fg="default", bg="default", options=None, align="left", cell_format=None
    ) -> None:
        self._fg = fg
        self._bg = bg
        self._options = options
        self._align = "left"
        self._cell_format = cell_format

    @property
    def cell_format(self) -> Optional[str]:
        return self._cell_format

    @property
    def tag(self) -> str:
        tag = "<fg={};bg={}"

        if self._options:
            tag += ";options={}".format(",".join(self._options))

        tag += ">"

        return tag

    def pad(self, string: str, length: int, char: str = " ") -> str:
        if self._align == "left":
            return string.rjust(length, char)

        if self._align == "right":
            return string.ljust(length, char)

        return string.center(length, char)



import ast
import inspect
import io
import keyword
import os
import re
import sys
import tokenize

from typing import List
from typing import Optional
from typing import Union

from crashtest.frame import Frame
from crashtest.frame_collection import FrameCollection
from crashtest.inspector import Inspector
from crashtest.solution_providers.solution_provider_repository import (
    SolutionProviderRepository,
)

from cleo.formatters.formatter import Formatter
from cleo.io.io import IO
from cleo.io.outputs.output import Output


class Highlighter:

    TOKEN_DEFAULT = "token_default"
    TOKEN_COMMENT = "token_comment"
    TOKEN_STRING = "token_string"
    TOKEN_NUMBER = "token_number"
    TOKEN_KEYWORD = "token_keyword"
    TOKEN_BUILTIN = "token_builtin"
    TOKEN_OP = "token_op"
    LINE_MARKER = "line_marker"
    LINE_NUMBER = "line_number"

    DEFAULT_THEME = {
        TOKEN_STRING: "fg=yellow;options=bold",
        TOKEN_NUMBER: "fg=blue;options=bold",
        TOKEN_COMMENT: "fg=default;options=dark,italic",
        TOKEN_KEYWORD: "fg=magenta;options=bold",
        TOKEN_BUILTIN: "fg=default;options=bold",
        TOKEN_DEFAULT: "fg=default",
        TOKEN_OP: "fg=default;options=dark",
        LINE_MARKER: "fg=red;options=bold",
        LINE_NUMBER: "fg=default;options=dark",
    }

    KEYWORDS = set(keyword.kwlist)
    BUILTINS = set(
        __builtins__.keys() if type(__builtins__) is dict else dir(__builtins__)
    )

    UI = {
        False: {"arrow": ">", "delimiter": "|"},
        True: {"arrow": "→", "delimiter": "│"},
    }

    def __init__(self, supports_utf8: bool = True) -> None:
        self._theme = self.DEFAULT_THEME.copy()
        self._ui = self.UI[supports_utf8]

    def code_snippet(
        self, source: str, line: int, lines_before: int = 2, lines_after: int = 2
    ) -> List[str]:
        token_lines = self.highlighted_lines(source)
        token_lines = self.line_numbers(token_lines, line)

        offset = line - lines_before - 1
        offset = max(offset, 0)
        length = lines_after + lines_before + 1
        token_lines = token_lines[offset : offset + length]

        return token_lines

    def highlighted_lines(self, source):
        source = source.replace("\r\n", "\n").replace("\r", "\n")

        return self.split_to_lines(source)

    def split_to_lines(self, source):
        lines = []
        current_line = 1
        current_col = 0
        buffer = ""
        current_type = None
        source_io = io.BytesIO(source.encode())
        formatter = Formatter()

        def readline():
            return formatter.format(
                formatter.escape(source_io.readline().decode())
            ).encode()

        tokens = tokenize.tokenize(readline)
        line = ""
        for token_info in tokens:
            token_type, token_string, start, end, _ = token_info
            lineno = start[0]
            if lineno == 0:
                # Encoding line
                continue

            if token_type == tokenize.ENDMARKER:
                # End of source
                if current_type is None:
                    current_type = self.TOKEN_DEFAULT

                line += "<{}>{}</>".format(self._theme[current_type], buffer)
                lines.append(line)
                break

            if lineno > current_line:
                if current_type is None:
                    current_type = self.TOKEN_DEFAULT

                diff = lineno - current_line
                if diff > 1:
                    lines += [""] * (diff - 1)

                line += "<{}>{}</>".format(
                    self._theme[current_type], buffer.rstrip("\n")
                )

                # New line
                lines.append(line)
                line = ""
                current_line = lineno
                current_col = 0
                buffer = ""

            if token_string in self.KEYWORDS:
                new_type = self.TOKEN_KEYWORD
            elif token_string in self.BUILTINS or token_string == "self":
                new_type = self.TOKEN_BUILTIN
            elif token_type == tokenize.STRING:
                new_type = self.TOKEN_STRING
            elif token_type == tokenize.NUMBER:
                new_type = self.TOKEN_NUMBER
            elif token_type == tokenize.COMMENT:
                new_type = self.TOKEN_COMMENT
            elif token_type == tokenize.OP:
                new_type = self.TOKEN_OP
            elif token_type == tokenize.NEWLINE:
                continue
            else:
                new_type = self.TOKEN_DEFAULT

            if current_type is None:
                current_type = new_type

            if start[1] > current_col:
                buffer += token_info.line[current_col : start[1]]

            if current_type != new_type:
                line += "<{}>{}</>".format(self._theme[current_type], buffer)
                buffer = ""
                current_type = new_type

            if lineno < end[0]:
                # The token spans multiple lines
                token_lines = token_string.split("\n")
                line += "<{}>{}</>".format(self._theme[current_type], token_lines[0])
                lines.append(line)
                for token_line in token_lines[1:-1]:
                    lines.append(
                        "<{}>{}</>".format(self._theme[current_type], token_line)
                    )

                current_line = end[0]
                buffer = token_lines[-1][: end[1]]
                line = ""
                continue

            buffer += token_string
            current_col = end[1]
            current_line = lineno

        return lines

    def line_numbers(
        self, lines: List[str], mark_line: Optional[int] = None
    ) -> List[str]:
        max_line_length = max(3, len(str(len(lines))))

        snippet_lines = []
        marker = "<{}>{}</> ".format(self._theme[self.LINE_MARKER], self._ui["arrow"])
        no_marker = "  "
        for i, line in enumerate(lines):
            if mark_line is not None:
                if mark_line == i + 1:
                    snippet = marker
                else:
                    snippet = no_marker

            line_number = "{:>{}}".format(i + 1, max_line_length)
            snippet += "<{}>{}</><{}>{}</> {}".format(
                "fg=default;options=bold"
                if mark_line == i + 1
                else self._theme[self.LINE_NUMBER],
                line_number,
                self._theme[self.LINE_NUMBER],
                self._ui["delimiter"],
                line,
            )
            snippet_lines.append(snippet)

        return snippet_lines


class ExceptionTrace:
    """
    Renders the trace of an exception.
    """

    THEME = {
        "comment": "<fg=black;options=bold>",
        "keyword": "<fg=yellow>",
        "builtin": "<fg=blue>",
        "literal": "<fg=magenta>",
    }

    AST_ELEMENTS = {
        "builtins": __builtins__.keys()
        if type(__builtins__) is dict
        else dir(__builtins__),
        "keywords": [
            getattr(ast, cls)
            for cls in dir(ast)
            if keyword.iskeyword(cls.lower())
            and inspect.isclass(getattr(ast, cls))
            and issubclass(getattr(ast, cls), ast.AST)
        ],
    }

    _FRAME_SNIPPET_CACHE = {}

    def __init__(
        self,
        exception: Exception,
        solution_provider_repository: Optional[SolutionProviderRepository] = None,
    ) -> None:
        self._exception = exception
        self._solution_provider_repository = solution_provider_repository
        self._exc_info = sys.exc_info()
        self._ignore = None

    def ignore_files_in(self, ignore: str) -> "ExceptionTrace":
        self._ignore = ignore

        return self

    def render(self, io: Union[IO, Output], simple: bool = False) -> None:
        if simple:
            io.write_line("")
            io.write_line("<error>{}</error>".format(str(self._exception)))
            return

        return self._render_exception(io, self._exception)

    def _render_exception(self, io: Union[IO, Output], exception: Exception) -> None:
        from crashtest.inspector import Inspector

        inspector = Inspector(exception)
        if not inspector.frames:
            return

        if inspector.has_previous_exception():
            self._render_exception(io, inspector.previous_exception)
            io.write_line("")
            io.write_line(
                "The following error occurred when trying to handle this error:"
            )
            io.write_line("")

        self._render_trace(io, inspector.frames)

        self._render_line(
            io, "<error>{}</error>".format(inspector.exception_name), True
        )
        io.write_line("")
        exception_message = (
            Formatter().format(inspector.exception_message).replace("\n", "\n  ")
        )
        self._render_line(io, "<b>{}</b>".format(exception_message))

        current_frame = inspector.frames[-1]
        self._render_snippet(io, current_frame)

        self._render_solution(io, inspector)

    def _render_snippet(self, io: Union[IO, Output], frame: Frame):
        self._render_line(
            io,
            "at <fg=green>{}</>:<b>{}</b> in <fg=cyan>{}</>".format(
                self._get_relative_file_path(frame.filename),
                frame.lineno,
                frame.function,
            ),
            True,
        )

        code_lines = Highlighter(supports_utf8=io.supports_utf8()).code_snippet(
            frame.file_content, frame.lineno, 4, 4
        )

        for code_line in code_lines:
            self._render_line(io, code_line, indent=4)

    def _render_solution(self, io: Union[IO, Output], inspector: Inspector):
        if self._solution_provider_repository is None:
            return

        solutions = self._solution_provider_repository.get_solutions_for_exception(
            inspector.exception
        )
        symbol = "•"
        if not io.supports_utf8():
            symbol = "*"

        for solution in solutions:
            title = solution.solution_title
            description = solution.solution_description
            links = solution.documentation_links

            description = description.replace("\n", "\n    ").strip(" ")

            self._render_line(
                io,
                "<fg=blue;options=bold>{} </><fg=default;options=bold>{}</>: {}{}".format(
                    symbol,
                    title.rstrip("."),
                    description,
                    ",".join("\n    <fg=blue>{}</>".format(link) for link in links),
                ),
                True,
            )

    def _render_trace(self, io: Union[IO, Output], frames: FrameCollection) -> None:
        stack_frames = FrameCollection()
        for frame in frames:
            if (
                self._ignore
                and re.match(self._ignore, frame.filename)
                and not io.is_debug()
            ):
                continue

            stack_frames.append(frame)

        remaining_frames_length = len(stack_frames) - 1
        if io.is_verbose() and remaining_frames_length:
            self._render_line(io, "<fg=yellow>Stack trace</>:", True)
            max_frame_length = len(str(remaining_frames_length))
            frame_collections = stack_frames.compact()
            i = remaining_frames_length
            for collection in frame_collections:
                if collection.is_repeated():
                    if len(collection) > 1:
                        frames_message = "<fg=yellow>{}</> frames".format(
                            len(collection)
                        )
                    else:
                        frames_message = "frame"

                    self._render_line(
                        io,
                        "<fg=blue>{:>{}}</>  Previous {} repeated <fg=blue>{}</> times".format(
                            "...",
                            max_frame_length,
                            frames_message,
                            collection.repetitions + 1,
                        ),
                        True,
                    )

                    i -= len(collection) * (collection.repetitions + 1)

                for frame in collection:
                    relative_file_path = self._get_relative_file_path(frame.filename)
                    relative_file_path_parts = relative_file_path.split(os.path.sep)
                    relative_file_path = "{}".format(
                        "<fg=default;options=dark>{}</>".format(
                            Formatter.escape(os.sep)
                        ).join(
                            relative_file_path_parts[:-1]
                            + [
                                "<fg=default;options=bold>{}</>".format(
                                    relative_file_path_parts[-1]
                                )
                            ]
                        ),
                    )

                    self._render_line(
                        io,
                        "<fg=yellow>{:>{}}</>  {}<fg=default;options=dark>:</><b>{}</b> in <fg=cyan>{}</>".format(
                            i,
                            max_frame_length,
                            relative_file_path,
                            frame.lineno,
                            frame.function,
                        ),
                        True,
                    )

                    if io.is_debug():
                        if (frame, 2, 2) not in self._FRAME_SNIPPET_CACHE:
                            code_lines = Highlighter(
                                supports_utf8=io.supports_utf8()
                            ).code_snippet(
                                frame.file_content,
                                frame.lineno,
                            )

                            self._FRAME_SNIPPET_CACHE[(frame, 2, 2)] = code_lines

                        code_lines = self._FRAME_SNIPPET_CACHE[(frame, 2, 2)]

                        for code_line in code_lines:
                            self._render_line(
                                io,
                                "{:>{}}{}".format(" ", max_frame_length, code_line),
                                indent=3,
                            )
                    else:
                        highlighter = Highlighter(supports_utf8=io.supports_utf8())
                        try:
                            code_line = highlighter.highlighted_lines(
                                frame.line.strip()
                            )[0]
                        except tokenize.TokenError:
                            code_line = frame.line.strip()

                        self._render_line(
                            io,
                            "{:>{}}    {}".format(
                                " ",
                                max_frame_length,
                                code_line,
                            ),
                        )

                    i -= 1

    def _render_line(
        self, io: Union[IO, Output], line: str, new_line: bool = False, indent: int = 2
    ) -> None:
        if new_line:
            io.write_line("")

        io.write_line("{}{}".format(indent * " ", line))

    def _get_relative_file_path(self, filepath: str) -> str:
        cwd = os.getcwd()

        if cwd:
            filepath = filepath.replace(cwd + os.path.sep, "")

        home = os.path.expanduser("~")
        if home:
            filepath = filepath.replace(home + os.path.sep, "~" + os.path.sep)

        return filepath



import sys

from typing import Any
from typing import List
from typing import Optional
from typing import Union

from cleo.exceptions import NoSuchOptionException
from cleo.exceptions import RuntimeException
from cleo.io.inputs.definition import Definition

from .input import Input


class ArgvInput(Input):
    """
    Represents an input coming from the command line.
    """

    def __init__(
        self, argv: Optional[List[str]] = None, definition: Optional[Definition] = None
    ) -> None:
        if argv is None:
            argv = sys.argv

        argv = argv[:]

        # Strip the application name
        try:
            self._script_name = argv.pop(0)
        except IndexError:
            self._script_name = None

        self._tokens = argv
        self._parsed = []

        super().__init__(definition=definition)

    @property
    def first_argument(self) -> Optional[str]:
        is_option = False

        for i, token in enumerate(self._tokens):
            if token and token[0] == "-":
                if token.find("=") != -1 or len(self._tokens) <= (i + 1):
                    continue

                # If it's a long option, consider that
                # everything after "--" is the option name.
                # Otherwise, use the last character
                # (if it's a short option set, only the last one
                # can take a value with space separator).
                if len(token) > 1 and token[1] == "-":
                    name = token[2:]
                else:
                    name = token[-1]

                if name not in self._options and not self._definition.has_shortcut(
                    name
                ):
                    # noop
                    continue

                if name not in self._options:
                    name = self._definition.shortcut_to_name(name)

                if name in self._options and self._tokens[i + 1] == self._options[name]:
                    is_option = True

                continue

            if is_option:
                is_option = False
                continue

            return token

        return

    @property
    def script_name(self) -> Optional[str]:
        return self._script_name

    def has_parameter_option(
        self, values: Union[str, List[str]], only_params: bool = False
    ) -> bool:
        """
        Returns true if the raw parameters (not parsed) contain a value.
        """
        if not isinstance(values, list):
            values = [values]

        for token in self._tokens:
            if only_params and token == "--":
                return False

            for value in values:
                # Options with values:
                # For long options, test for '--option=' at beginning
                # For short options, test for '-o' at beginning
                if value.find("--") == 0:
                    leading = value + "="
                else:
                    leading = value

                if token == value or leading != "" and token.find(leading) == 0:
                    return True

        return False

    def parameter_option(
        self,
        values: Union[str, List[str]],
        default: Any = False,
        only_params: bool = False,
    ) -> Any:
        if not isinstance(values, list):
            values = [values]

        tokens = self._tokens[:]
        while len(tokens) > 0:
            token = tokens.pop(0)
            if only_params and token == "--":
                return default

            for value in values:
                if token == value:
                    try:
                        return tokens.pop(0)
                    except IndexError:
                        return

                # Options with values:
                # For long options, test for '--option=' at beginning
                # For short options, test for '-o' at beginning
                if value.find("--") == 0:
                    leading = value + "="
                else:
                    leading = value

                if token == value or leading != "" and token.find(leading) == 0:
                    return token[len(leading)]

        return False

    def _set_tokens(self, tokens: List[str]) -> None:
        self._tokens = tokens

    def _parse(self) -> None:
        parse_options = True
        self._parsed = self._tokens[:]

        try:
            token = self._parsed.pop(0)
        except IndexError:
            token = None

        while token is not None:
            if parse_options and token == "":
                self._parse_argument(token)
            elif parse_options and token == "--":
                parse_options = False
            elif parse_options and token.find("--") == 0:
                self._parse_long_option(token)
            elif parse_options and token[0] == "-" and token != "-":
                self._parse_short_option(token)
            else:
                self._parse_argument(token)

            try:
                token = self._parsed.pop(0)
            except IndexError:
                token = None

    def _parse_short_option(self, token: str) -> None:
        name = token[1:]

        if len(name) > 1:
            if (
                self._definition.has_shortcut(name[0])
                and self._definition.option_for_shortcut(name[0]).accepts_value()
            ):
                # An option with a value and no space
                self._add_short_option(name[0], name[1:])
            else:
                self._parse_short_option_set(name)
        else:
            self._add_short_option(name, None)

    def _parse_short_option_set(self, name: str) -> None:
        length = len(name)
        for i in range(length):
            if not self._definition.has_shortcut(name[i]):
                raise RuntimeException(f'The option "{name[i]}" does not exist')

            option = self._definition.option_for_shortcut(name[i])
            if option.accepts_value():
                self._add_long_option(
                    option.name, None if i == length - 1 else name[i + 1 :]
                )

                break

            self._add_long_option(option.name, None)

    def _parse_long_option(self, token: str) -> None:
        name = token[2:]

        pos = name.find("=")
        if pos != -1:
            value = name[pos + 1 :]
            if not value:
                self._parsed.insert(0, value)

            self._add_long_option(name[0:pos], value)
        else:
            self._add_long_option(name, None)

    def _parse_argument(self, token: str) -> None:
        count = len(self._arguments)

        # If the input is expecting another argument, add it
        if self._definition.has_argument(count):
            argument = self._definition.argument(count)
            self._arguments[argument.name] = [token] if argument.is_list() else token
        # If the last argument is a list, append the token to it
        elif (
            self._definition.has_argument(count - 1)
            and self._definition.argument(count - 1).is_list()
        ):
            argument = self._definition.argument(count - 1)
            self._arguments[argument.name].append(token)
        # Unexpected argument
        else:
            all = self._definition.arguments.copy()
            command_name = None
            argument = all[0]
            if argument and argument.name == "command":
                command_name = self._arguments.get("command")
                del all[0]

            if all:
                if command_name:
                    message = 'Too many arguments to "{}" command, expected arguments "{}"'.format(
                        command_name, '" "'.join([a.name for a in all])
                    )
                else:
                    message = 'Too many arguments, expected arguments "{}"'.format(
                        '" "'.join([a.name for a in all])
                    )
            elif command_name:
                message = 'No arguments expected for "{}" command, got "{}"'.format(
                    command_name, token
                )
            else:
                message = 'No arguments expected, got "{}"'.format(token)

            raise RuntimeException(message)

    def _add_short_option(self, shortcut: str, value: Any) -> None:
        if not self._definition.has_shortcut(shortcut):
            raise NoSuchOptionException(f'The option "-{shortcut}" does not exist')

        self._add_long_option(
            self._definition.option_for_shortcut(shortcut).name, value
        )

    def _add_long_option(self, name: str, value: Any) -> None:
        if not self._definition.has_option(name):
            raise NoSuchOptionException(f'The option "--{name}" does not exist')

        option = self._definition.option(name)

        if value is not None and not option.accepts_value():
            raise RuntimeException(f'The "--{name}" option does not accept a value')

        if value in ["", None] and option.accepts_value() and self._parsed:
            # If the option accepts a value, either required or optional,
            # we check if there is one
            next_token = self._parsed.pop(0)
            if (next_token and next_token[0] != "-") or next_token in ["", None]:
                value = next_token
            else:
                self._parsed.insert(0, next_token)

        if value is None:
            if option.requires_value():
                raise RuntimeException(f'The "--{name}" option requires a value')

            if not option.is_list() and option.is_flag():
                value = True

        if option.is_list():
            if name not in self._options:
                self._options[name] = []

            self._options[name].append(value)
        else:
            self._options[name] = value



import os
import platform
import shlex
import struct
import subprocess

from typing import Optional
from typing import Tuple


class Terminal:
    """
    Represents the current terminal.
    """

    _width = None
    _height = None

    @property
    def width(self) -> int:
        width = os.getenv("COLUMNS", "").strip()
        if width:
            return int(width)

        if self.__class__._width is None:
            self._init_dimensions()

        return self.__class__._width

    @property
    def height(self) -> int:
        height = os.getenv("LINES", "").strip()
        if height:
            return int(height)

        if self.__class__._height is None:
            self._init_dimensions()

        return self.__class__._height

    @classmethod
    def _init_dimensions(cls) -> None:
        current_os = platform.system().lower()
        dimensions = None

        if current_os.lower() == "windows":
            dimensions = cls._get_terminal_size_windows()
            if dimensions is None:
                dimensions = cls._get_terminal_size_tput()
        elif current_os.lower() in ["linux", "darwin"] or current_os.startswith(
            "cygwin"
        ):
            dimensions = cls._get_terminal_size_linux()

        if dimensions is None:
            dimensions = 80, 25

        # Ensure we have a valid width
        if dimensions[0] <= 0:
            dimensions = 80, dimensions[1]

        cls._width, cls._height = dimensions

    @classmethod
    def _get_terminal_size_windows(cls) -> Optional[Tuple[int, int]]:
        try:
            from ctypes import create_string_buffer
            from ctypes import windll

            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (
                    bufx,
                    bufy,
                    curx,
                    cury,
                    wattr,
                    left,
                    top,
                    right,
                    bottom,
                    maxx,
                    maxy,
                ) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                sizey = bottom - top + 1
                return sizex, sizey
        except Exception:
            return

    @classmethod
    def _get_terminal_size_tput(cls) -> Optional[Tuple[int, int]]:
        # get terminal width
        # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
        try:
            cols = int(
                subprocess.check_output(
                    shlex.split("tput cols"), stderr=subprocess.STDOUT
                )
            )
            rows = int(
                subprocess.check_output(
                    shlex.split("tput lines"), stderr=subprocess.STDOUT
                )
            )

            return cols, rows
        except Exception:
            pass

    @classmethod
    def _get_terminal_size_linux(cls) -> Optional[Tuple[int, int]]:
        def ioctl_GWINSZ(fd):
            try:
                import fcntl
                import termios

                cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
                return cr
            except Exception:
                pass

        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)
            except Exception:
                pass

        if not cr:
            try:
                cr = (os.environ["LINES"], os.environ["COLUMNS"])
            except Exception:
                return

        return int(cr[1]), int(cr[0])



from typing import List
from typing import Optional

from cleo._utils import find_similar_names


class CleoException(Exception):

    exit_code: Optional[int] = None


class CleoSimpleException(Exception):

    pass


class LogicException(CleoException):

    pass


class RuntimeException(CleoException):

    pass


class ValueException(CleoException):

    pass


class MissingArgumentsException(CleoSimpleException):

    pass


class NoSuchOptionException(CleoException):

    pass


class CommandNotFoundException(CleoSimpleException):
    def __init__(self, name: str, commands: Optional[List[str]] = None) -> None:
        message = 'The command "{}" does not exist.'.format(name)

        if commands:
            suggested_names = find_similar_names(name, commands)

            if suggested_names:
                if len(suggested_names) == 1:
                    message += "\n\nDid you mean this?\n    "
                else:
                    message += "\n\nDid you mean one of these?\n    "

                message += "\n    ".join(suggested_names)

        super().__init__(message)


class NamespaceNotFoundException(CleoSimpleException):
    def __init__(self, name: str, namespaces: Optional[List[str]] = None) -> None:
        message = 'There are no commands in the "{}" namespace.'.format(name)

        if namespaces:
            suggested_names = find_similar_names(name, namespaces)

            if suggested_names:
                if len(suggested_names) == 1:
                    message += "\n\nDid you mean this?\n    "
                else:
                    message += "\n\nDid you mean one of these?\n    "

                message += "\n    ".join(suggested_names)

        super().__init__(message)



import math

from html.parser import HTMLParser
from typing import List

from pylev import levenshtein


class TagStripper(HTMLParser):
    def __init__(self) -> None:
        super().__init__(convert_charrefs=False)

        self.reset()
        self.fed = []

    def handle_data(self, d) -> None:
        self.fed.append(d)

    def handle_entityref(self, name) -> None:
        self.fed.append("&%s;" % name)

    def handle_charref(self, name) -> None:
        self.fed.append("&#%s;" % name)

    def get_data(self) -> str:
        return "".join(self.fed)


def _strip(value) -> str:
    s = TagStripper()
    s.feed(value)
    s.close()

    return s.get_data()


def strip_tags(value):
    value = str(value)
    while "<" in value and ">" in value:
        new_value = _strip(value)
        if value.count("<") == new_value.count("<"):
            break

        value = new_value

    return value


def find_similar_names(name, names):  # type: (str, List[str]) -> List[str]
    """
    Finds names similar to a given command name.
    """
    threshold = 1e3
    distance_by_name = {}
    suggested_names = []

    for actual_name in names:
        # Get Levenshtein distance between the input and each command name
        distance = levenshtein(name, actual_name)

        is_similar = distance <= len(name) / 3
        is_sub_string = actual_name.find(name) != -1

        if is_similar or is_sub_string:
            distance_by_name[actual_name] = (
                distance,
                actual_name.find(name) if is_sub_string else float("inf"),
            )

    # Only keep results with a distance below the threshold
    distance_by_name = {
        k: v for k, v in distance_by_name.items() if v[0] < 2 * threshold
    }

    # Display results with shortest distance first
    for k, v in sorted(distance_by_name.items(), key=lambda i: (i[1][0], i[1][1])):
        if k not in suggested_names:
            suggested_names.append(k)

    return suggested_names


_TIME_FORMATS = [
    (0, "< 1 sec"),
    (2, "1 sec"),
    (59, "secs", 1),
    (60, "1 min"),
    (3600, "mins", 60),
    (5400, "1 hr"),
    (86400, "hrs", 3600),
    (129600, "1 day"),
    (604800, "days", 86400),
]


def format_time(secs):  # type: (float) -> str
    for fmt in _TIME_FORMATS:
        if secs > fmt[0]:
            continue

        if len(fmt) == 2:
            return fmt[1]

        return "{} {}".format(math.ceil(secs / fmt[2]), fmt[1])



from typing import Iterable
from typing import Optional
from typing import Union

from .inputs.input import Input
from .outputs.output import Output
from .outputs.output import Type as OutputType
from .outputs.output import Verbosity
from .outputs.section_output import SectionOutput


class IO:
    def __init__(self, input: Input, output: Output, error_output: Output) -> None:
        self._input = input
        self._output = output
        self._error_output = error_output

    @property
    def input(self) -> Input:
        return self._input

    @property
    def output(self) -> Output:
        return self._output

    @property
    def error_output(self) -> Output:
        return self._error_output

    def read(self, length: int, default: Optional[str] = None) -> str:
        """
        Reads the given amount of characters from the input stream.
        """
        return self._input.read(length, default=default)

    def read_line(
        self, length: Optional[int] = None, default: Optional[str] = None
    ) -> str:
        """
        Reads a line from the input stream.
        """
        return self._input.read_line(length=length, default=default)

    def write_line(
        self,
        messages: Union[str, Iterable[str]],
        verbosity: Verbosity = Verbosity.NORMAL,
        type: OutputType = OutputType.NORMAL,
    ) -> None:
        self._output.write_line(messages, verbosity=verbosity, type=type)

    def write(
        self,
        messages: Union[str, Iterable[str]],
        new_line: bool = False,
        verbosity: Verbosity = Verbosity.NORMAL,
        type: OutputType = OutputType.NORMAL,
    ) -> None:
        self._output.write(messages, new_line=new_line, verbosity=verbosity, type=type)

    def write_error_line(
        self,
        messages: Union[str, Iterable[str]],
        verbosity: Verbosity = Verbosity.NORMAL,
        type: OutputType = OutputType.NORMAL,
    ) -> None:
        self._error_output.write_line(messages, verbosity=verbosity, type=type)

    def write_error(
        self,
        messages: Union[str, Iterable[str]],
        new_line: bool = False,
        verbosity: Verbosity = Verbosity.NORMAL,
        type: OutputType = OutputType.NORMAL,
    ) -> None:
        self._error_output.write(
            messages, new_line=new_line, verbosity=verbosity, type=type
        )

    def overwrite(self, messages: Union[str, Iterable[str]]) -> None:
        from cleo.cursor import Cursor

        cursor = Cursor(self._output)
        cursor.move_to_column(1)
        cursor.clear_line()
        self.write(messages)

    def overwrite_error(self, messages: Union[str, Iterable[str]]) -> None:
        from cleo.cursor import Cursor

        cursor = Cursor(self._error_output)
        cursor.move_to_column(1)
        cursor.clear_line()
        self.write_error(messages)

    def flush(self) -> None:
        self._output.flush()

    def is_interactive(self) -> bool:
        return self._input.is_interactive()

    def interactive(self, interactive: bool = True) -> None:
        self._input.interactive(interactive)

    def decorated(self, decorated: bool = True) -> None:
        self._output.decorated(decorated)
        self._error_output.decorated(decorated)

    def is_decorated(self) -> bool:
        return self._output.is_decorated()

    def supports_utf8(self) -> bool:
        return self._output.supports_utf8()

    def set_verbosity(self, verbosity: Verbosity) -> None:
        self._output.set_verbosity(verbosity)
        self._error_output.set_verbosity(verbosity)

    def is_verbose(self) -> bool:
        return self.output.is_verbose()

    def is_very_verbose(self) -> bool:
        return self.output.is_very_verbose()

    def is_debug(self) -> bool:
        return self.output.is_debug()

    def set_input(self, input: Input) -> None:
        self._input = input

    def with_input(self, input: Input) -> "IO":
        return self.__class__(input, self._output, self._error_output)

    def remove_format(self, text: str) -> str:
        return self._output.remove_format(text)

    def section(self) -> SectionOutput:
        return self._output.section()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boeah.cleo',
 'boeah.cleo.commands',
 'boeah.cleo.commands.completions',
 'boeah.cleo.descriptors',
 'boeah.cleo.events',
 'boeah.cleo.exceptions',
 'boeah.cleo.formatters',
 'boeah.cleo.io',
 'boeah.cleo.io.inputs',
 'boeah.cleo.io.outputs',
 'boeah.cleo.loaders',
 'boeah.cleo.testers',
 'boeah.cleo.ui']

package_data = \
{'': ['*']}

install_requires = \
['crashtest>=0.3.1,<0.4.0', 'pylev>=1.3.0,<2.0.0']

setup_kwargs = {
    'name': 'boeah.cleo',
    'version': '1.0.0',
    'description': 'Cleo allows you to create beautiful and testable command-line interfaces.',
    'long_description': 'Cleo\n####\n\n.. image:: https://travis-ci.org/sdispater/cleo.png\n   :alt: Cleo Build status\n   :target: https://travis-ci.org/sdispater/cleo\n\nCreate beautiful and testable command-line interfaces.\n\nCleo is mostly a higher level wrapper for `CliKit <https://github.com/sdispater/clikit>`_, so\na lot of the components and utilities comes from it. Refer to its documentation for more\ninformation.\n\nResources\n=========\n\n* `Documentation <http://cleo.readthedocs.io>`_\n* `Issue Tracker <https://github.com/sdispater/cleo/issues>`_\n\n\nUsage\n=====\n\nTo make a command that greets you from the command line,\ncreate ``greet_command.py`` and add the following to it:\n\n.. code-block:: python\n\n    from cleo import Command\n\n\n    class GreetCommand(Command):\n        """\n        Greets someone\n\n        greet\n            {name? : Who do you want to greet?}\n            {--y|yell : If set, the task will yell in uppercase letters}\n        """\n\n        def handle(self):\n            name = self.argument(\'name\')\n\n            if name:\n                text = \'Hello {}\'.format(name)\n            else:\n                text = \'Hello\'\n\n            if self.option(\'yell\'):\n                text = text.upper()\n\n            self.line(text)\n\n\nYou also need to create the file to run at the command line which creates\nan ``Application`` and adds commands to it:\n\n.. code-block:: python\n\n    #!/usr/bin/env python\n\n    from greet_command import GreetCommand\n    from cleo import Application\n\n    application = Application()\n    application.add(GreetCommand())\n\n    if __name__ == \'__main__\':\n        application.run()\n\nTest the new command by running the following\n\n.. code-block:: bash\n\n    $ python application.py greet John\n\nThis will print the following to the command line:\n\n.. code-block:: text\n\n    Hello John\n\nYou can also use the ``--yell`` option to make everything uppercase:\n\n.. code-block:: bash\n\n    $ python application.py greet John --yell\n\nThis prints:\n\n.. code-block:: text\n\n    HELLO JOHN\n\nAs you may have already seen, Cleo uses the command docstring to determine\nthe command definition.\nThe docstring must be in the following form :\n\n.. code-block:: python\n\n    """\n    Command description\n\n    Command signature\n    """\n\nThe signature being in the following form:\n\n.. code-block:: python\n\n    """\n    command:name {argument : Argument description} {--option : Option description}\n    """\n\nThe signature can span multiple lines.\n\n.. code-block:: python\n\n    """\n    command:name\n        {argument : Argument description}\n        {--option : Option description}\n    """\n\nColoring the Output\n-------------------\n\nWhenever you output text, you can surround the text with tags to color its\noutput. For example:\n\n.. code-block:: python\n\n    # green text\n    self.line(\'<info>foo</info>\')\n\n    # yellow text\n    self.line(\'<comment>foo</comment>\')\n\n    # black text on a cyan background\n    self.line(\'<question>foo</question>\')\n\n    # white text on a red background\n    self.line(\'<error>foo</error>\')\n\nThe closing tag can be replaced by ``</>``, which revokes all formatting options established by the last opened tag.\n\nIt is possible to define your own styles using the ``add_style()`` method:\n\n.. code-block:: python\n\n    self.add_style(\'fire\', fg=\'red\', bg=\'yellow\', options=[\'bold\', \'blink\'])\n    self.line(\'<fire>foo</fire>\')\n\nAvailable foreground and background colors are: ``black``, ``red``, ``green``,\n``yellow``, ``blue``, ``magenta``, ``cyan`` and ``white``.\n\nAnd available options are: ``bold``, ``underscore``, ``blink``, ``reverse`` and ``conceal``.\n\nYou can also set these colors and options inside the tag name:\n\n.. code-block:: python\n\n    # green text\n    self.line(\'<fg=green>foo</>\')\n\n    # black text on a cyan background\n    self.line(\'<fg=black;bg=cyan>foo</>\')\n\n    # bold text on a yellow background\n    self.line(\'<bg=yellow;options=bold>foo</>\')\n\n\nVerbosity Levels\n----------------\n\nCleo has four verbosity levels. These are defined in the ``Output`` class:\n\n=======================================  ================================== ======================\nMode                                     Meaning                            Console option\n=======================================  ================================== ======================\n``NA``                                   Do not output any messages         ``-q`` or ``--quiet``\n``clikit.VERBOSITY_NORMAL``              The default verbosity level        (none)\n``clikit.VERBOSITY_VERBOSE``             Increased verbosity of messages    ``-v``\n``clikit.VERBOSITY_VERY_VERBOSE``        Informative non essential messages ``-vv``\n``clikit.VERBOSITY_DEBUG``               Debug messages                     ``-vvv``\n=======================================  ================================== ======================\n\nIt is possible to print a message in a command for only a specific verbosity\nlevel. For example:\n\n.. code-block:: python\n\n    if clikit.VERBOSITY_VERBOSE <= self.io.verbosity:\n        self.line(...)\n\nThere are also more semantic methods you can use to test for each of the\nverbosity levels:\n\n.. code-block:: python\n\n    if self.output.is_quiet():\n        # ...\n\n    if self.output.is_verbose():\n        # ...\n\nYou can also pass the verbosity flag directly to `line()`.\n\n.. code-block:: python\n\n    self.line("", verbosity=clikit.VERBOSITY_VERBOSE)\n\nWhen the quiet level is used, all output is suppressed.\n\n\nUsing Arguments\n---------------\n\nThe most interesting part of the commands are the arguments and options that\nyou can make available. Arguments are the strings - separated by spaces - that\ncome after the command name itself. They are ordered, and can be optional\nor required. For example, add an optional ``last_name`` argument to the command\nand make the ``name`` argument required:\n\n.. code-block:: python\n\n    class GreetCommand(Command):\n        """\n        Greets someone\n\n        greet\n            {name : Who do you want to greet?}\n            {last_name? : Your last name?}\n            {--y|yell : If set, the task will yell in uppercase letters}\n        """\n\nYou now have access to a ``last_name`` argument in your command:\n\n.. code-block:: python\n\n    last_name = self.argument(\'last_name\')\n    if last_name:\n        text += \' {}\'.format(last_name)\n\nThe command can now be used in either of the following ways:\n\n.. code-block:: bash\n\n    $ python application.py greet John\n    $ python application.py greet John Doe\n\nIt is also possible to let an argument take a list of values (imagine you want\nto greet all your friends). For this it must be specified at the end of the\nargument list:\n\n.. code-block:: python\n\n    class GreetCommand(Command):\n        """\n        Greets someone\n\n        greet\n            {names* : Who do you want to greet?}\n            {--y|yell : If set, the task will yell in uppercase letters}\n        """\n\nTo use this, just specify as many names as you want:\n\n.. code-block:: bash\n\n    $ python application.py demo:greet John Jane\n\nYou can access the ``names`` argument as a list:\n\n.. code-block:: python\n\n    names = self.argument(\'names\')\n    if names:\n        text += \' {}\'.format(\', \'.join(names))\n\nThere are 3 argument variants you can use:\n\n================================ ==================================== ===============================================================================================================\nMode                             Notation                             Value\n================================ ==================================== ===============================================================================================================\n``clikit.ARGUMENT_REQUIRED``     none (just write the argument name)  The argument is required\n``clikit.ARGUMENT_OPTIONAL``     ``argument?``                        The argument is optional and therefore can be omitted\n``clikit.ARGUMENT_MULTI_VALUED`` ``argument*``                        The argument can contain an indefinite number of arguments and must be used at the end of the argument list\n================================ ==================================== ===============================================================================================================\n\nYou can combine them like this:\n\n.. code-block:: python\n\n    class GreetCommand(Command):\n        """\n        Greets someone\n\n        greet\n            {names?* : Who do you want to greet?}\n            {--y|yell : If set, the task will yell in uppercase letters}\n        """\n\nIf you want to set a default value, you can it like so:\n\n.. code-block:: text\n\n    argument=default\n\nThe argument will then be considered optional.\n\n\nUsing Options\n-------------\n\nUnlike arguments, options are not ordered (meaning you can specify them in any\norder) and are specified with two dashes (e.g. ``--yell`` - you can also\ndeclare a one-letter shortcut that you can call with a single dash like\n``-y``). Options are *always* optional, and can be setup to accept a value\n(e.g. ``--dir=src``) or simply as a boolean flag without a value (e.g.\n``--yell``).\n\n.. tip::\n\n    It is also possible to make an option *optionally* accept a value (so that\n    ``--yell`` or ``--yell=loud`` work). Options can also be configured to\n    accept a list of values.\n\nFor example, add a new option to the command that can be used to specify\nhow many times in a row the message should be printed:\n\n.. code-block:: python\n\n    class GreetCommand(Command):\n        """\n        Greets someone\n\n        greet\n            {name? : Who do you want to greet?}\n            {--y|yell : If set, the task will yell in uppercase letters}\n            {--iterations=1 : How many times should the message be printed?}\n        """\n\n\nNext, use this in the command to print the message multiple times:\n\n.. code-block:: python\n\n    for _ in range(0, self.option(\'iterations\')):\n        self.line(text)\n\nNow, when you run the task, you can optionally specify a ``--iterations``\nflag:\n\n.. code-block:: bash\n\n    $ python application.py demo:greet John\n    $ python application.py demo:greet John --iterations=5\n\nThe first example will only print once, since ``iterations`` is empty and\ndefaults to ``1``. The second example will print five times.\n\nRecall that options don\'t care about their order. So, either of the following\nwill work:\n\n.. code-block:: bash\n\n    $ python application.py demo:greet John --iterations=5 --yell\n    $ python application.py demo:greet John --yell --iterations=5\n\nThere are 4 option variants you can use:\n\n================================  =================================== ======================================================================================\nOption                            Notation                            Value\n================================  =================================== ======================================================================================\n``clikit.OPTION_MULTI_VALUED``    ``--option=*``                      This option accepts multiple values (e.g. ``--dir=/foo --dir=/bar``)\n``clikit.OPTION_NO_VALUE``        ``--option``                        Do not accept input for this option (e.g. ``--yell``)\n``clikit.OPTION_REQUIRED_VALUE``  ``--option=``                       This value is required (e.g. ``--iterations=5``), the option itself is still optional\n``clikit.OPTION_OPTIONAL_VALUE``  ``--option=?``                      This option may or may not have a value (e.g. ``--yell`` or ``--yell=loud``)\n================================  =================================== ======================================================================================\n\nYou can combine them like this:\n\n.. code-block:: python\n\n    class GreetCommand(Command):\n        """\n        Greets someone\n\n        greet\n            {name? : Who do you want to greet?}\n            {--y|yell : If set, the task will yell in uppercase letters}\n            {--iterations=?*1 : How many times should the message be printed?}\n        """\n\n\nTesting Commands\n----------------\n\nCleo provides several tools to help you test your commands. The most\nuseful one is the ``CommandTester`` class.\nIt uses a special IO class to ease testing without a real\nconsole:\n\n.. code-block:: python\n\n    import pytest\n\n    from cleo import Application\n    from cleo import CommandTester\n\n    def test_execute(self):\n        application = Application()\n        application.add(GreetCommand())\n\n        command = application.find(\'demo:greet\')\n        command_tester = CommandTester(command)\n        command_tester.execute()\n\n        assert "..." == tester.io.fetch_output()\n\nThe ``CommandTester.io.fetch_output()`` method returns what would have been displayed\nduring a normal call from the console. ``CommandTester.io.fetch_error()`` is also available\nto get what you have been written to the stderr.\n\nYou can test sending arguments and options to the command by passing them\nas a string to the ``CommandTester.execute()`` method:\n\n.. code-block:: python\n\n    import pytest\n\n    from cleo import Application\n    from cleo import CommandTester\n\n    def test_execute(self):\n        application = Application()\n        application.add(GreetCommand())\n\n        command = application.find(\'demo:greet\')\n        command_tester = CommandTester(command)\n        command_tester.execute("John")\n\n        assert "John" in tester.io.fetch_output()\n\nYou can also test a whole console application by using the ``ApplicationTester`` class.\n\n\nCalling an existing Command\n---------------------------\n\nIf a command depends on another one being run before it, instead of asking the\nuser to remember the order of execution, you can call it directly yourself.\nThis is also useful if you want to create a "meta" command that just runs a\nbunch of other commands.\n\nCalling a command from another one is straightforward:\n\n.. code-block:: python\n\n    def handle(self):\n        return_code = self.call(\'demo:greet\', "John --yell")\n\n        # ...\n\nIf you want to suppress the output of the executed command,\nyou can use the ``call_silent()`` method instead.\n\n\n\nAutocompletion\n--------------\n\nCleo supports automatic (tab) completion in ``bash``, ``zsh`` and ``fish``.\n\nTo activate support for autocompletion, pass a ``complete`` keyword when initializing\nyour application:\n\n.. code-block:: python\n\n    application = Application(\'My Application\', \'0.1\', complete=True)\n\nNow, register completion for your application by running one of the following in a terminal,\nreplacing ``[program]`` with the command you use to run your application:\n\n.. code-block:: bash\n\n    # BASH - Ubuntu / Debian\n    [program] completions bash | sudo tee /etc/bash_completion.d/[program].bash-completion\n\n    # BASH - Mac OSX (with Homebrew "bash-completion")\n    [program] completions bash > $(brew --prefix)/etc/bash_completion.d/[program].bash-completion\n\n    # ZSH - Config file\n    mkdir ~/.zfunc\n    echo "fpath+=~/.zfunc" >> ~/.zshrc\n    [program] completions zsh > ~/.zfunc/_test\n\n    # FISH\n    [program] completions fish > ~/.config/fish/completions/[program].fish\n',
    'author': 'Sébastien Eustace',
    'author_email': 'sebastien@eustace.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdispater/cleo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
